/* Audio Library for Teensy 3.X
   Copyright (c) 2014, Pete (El Supremo)
   Copyright (c) 2019, Holger Wirtz

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
   THE SOFTWARE.
*/

#include <Arduino.h>
#include "limits.h"
#include "effect_modulated_chorus.h"

/******************************************************************/

// Based on;      A u d i o E f f e c t C h o r u s
// Written by Pete (El Supremo) Jan 2014
// 140529 - change to handle mono stream - change modify() to voices()
// 140219 - correct storage class (not static)
// 190527 - adding modulation input (by Holger Wirtz)

boolean AudioModulatedEffectChorus::begin(short *delayline, int d_length)
{
#if 0
  Serial.print("AudioModulatedEffectChorus.begin(Chorus delay line length = ");
  Serial.print(d_length);
  Serial.println(")");
#endif

  l_delayline = NULL;
  delay_length = 0;
  l_circ_idx = 0;

  if (delayline == NULL) {
    return (false);
  }
  if (d_length < 10) {
    return (false);
  }

  l_delayline = delayline;
  delay_length = d_length;
  delay_length_half = d_length / 2;

  return (true);
}

//int last_idx = 0;
void AudioModulatedEffectChorus::update(void)
{
  audio_block_t *block;
  audio_block_t *modulation;
  short *bp;
  short *mp;
  float mod_idx;
  int x1;
  int x2;
  int y1;
  int y2;

  if (l_delayline == NULL)
    return;

  /*
    block = receiveWritable(0);
    if (block)
    {
      bp = block->data;
      uint32_t tmp = delay_length / (num_chorus - 1) - 1;
      for (int i = 0; i < AUDIO_BLOCK_SAMPLES; i++)
      {
        l_circ_idx++;
        if (l_circ_idx >= delay_length)
        {
          l_circ_idx = 0;
        }
        l_delayline[l_circ_idx] = *bp;
        sum = 0;
        c_idx = l_circ_idx;
        for (int k = 0; k < num_chorus; k++)
        {
          sum += l_delayline[c_idx];
          if (num_chorus > 1)c_idx -= tmp;
          if (c_idx < 0)
          {
            c_idx += delay_length;
          }
        }
        bp++ = sum / num_chorus;
      }
      // transmit the block
      transmit(block, 0);
      release(block);
    }
  */

  block = receiveWritable(0);
  modulation = receiveReadOnly(1);

  if (block && modulation)
  {
    bp = block->data;
    mp = modulation->data;

    for (int i = 0; i < AUDIO_BLOCK_SAMPLES; i++)
    {
      if (l_circ_idx >= delay_length)
        l_circ_idx = 0;
      l_delayline[l_circ_idx] = *bp; // write signal into circular buffer

      mod_idx = float(*mp) / SHRT_MAX * delay_length_half + l_circ_idx; // calculate index with modulation as a float

      if (mod_idx > delay_length)
        mod_idx = mod_idx - delay_length;
      else if (mod_idx < 0)
        mod_idx = delay_length + mod_idx;

      // linear interpolation
      x1 = int(mod_idx);
      y1 = l_delayline[x1];
      if (x1 + 1 >= delay_length)
        x2 = 0;
      else
        x2 = x1 + 1;
      y2 = l_delayline[x2];
      
      *bp = (int((float(y2 - y1) / (x2 - x1) * (mod_idx - x1) + y1) + 0.5) >> 1); // mix original signal 1:1 with modulated signal
      bp++;
      mp++;
      l_circ_idx++;
    }

    // transmit the block
    transmit(block, 0);
    release(block);
    release(modulation);
  }
}
